use salvo::{
    fs::NamedFile,
    handler,
    http::HeaderValue,
    hyper::header::{self, HOST},
    prelude::StatusCode,
    writer::Text,
    Request, Response,
};
use std::time::SystemTime;

use crate::{
    db::{self, FileCheckResult},
    engine,
    handlers::{render_template, TemplateStruct},
    CONFIG, SQLITE,
};

/// This handles serving files from disk, and all the tracking that goes with it.
#[handler]
pub async fn serve_file(req: &mut Request, res: &mut Response) {
    let headers = req.headers().clone();
    let sqlconn = SQLITE.get().unwrap();

    // Check if the filename exists in the DB
    let filename: String = req.param("file").unwrap();
    let check_filename = db::check_filename(sqlconn, &filename).await;

    // Returns true if the "source" field in the HTTP request's form data is "web", false otherwise or if an error occurs.
    // Since we don't have a 'source' field in the request, we guess from the user agent. It's not perfect, but it works. and it's not really a security problem.
    // Like worst case, some wacky browser ends up seeing the json api instead of the main page.
    // Well okay this might mess with crawling or something but meh.
    // Check if the request came from a web browser or something else.
    let user_agent = headers
        .get(header::USER_AGENT)
        .and_then(|value| value.to_str().ok())
        .unwrap_or("");
    let upload_by_web = user_agent.contains("Mozilla");

    match check_filename {
        Ok(FileCheckResult::NotFound) => {
            // If the file isn't valid, show a nice 404
            if upload_by_web {
                let template_filename = "error.html";

                let template = TemplateStruct {
                    domain: String::from(headers[HOST].to_str().unwrap()),
                    message1: String::from("Error 404: File not Found"),
                    message2: String::from(
                        "We couldn't find that. Are you sure you know what you're looking for?",
                    ),
                    ..Default::default()
                };

                let status_code = StatusCode::NOT_FOUND;
                render_template(res, &headers, template_filename, template, status_code).await;
            } else {
                res.status_code(StatusCode::NOT_FOUND);
                res.render(Text::Json(r#"{"error": "FileNotFound"}"#));
            }

            tracing::info!("File could not be found: {}", &filename);
        }

        Err(err) => {
            // Something went very wrong while running the query, return a error!
            tracing::error!("Error while trying to check the filename: {:?}", err);

            if upload_by_web {
                let template_filename = "error.html";

                let template = TemplateStruct {
                    domain: String::from(headers[HOST].to_str().unwrap()),
                    message1: String::from("Error 500: Internal Server Error"),
                    message2: String::from(
                        "Something went terribly wrong while trying to discover the file",
                    ),
                    ..Default::default()
                };

                let status_code = StatusCode::INTERNAL_SERVER_ERROR;
                render_template(res, &headers, template_filename, template, status_code).await;
            } else {
                res.status_code(StatusCode::UNAUTHORIZED);
                res.render(Text::Json(r#"{"error": "InternalServerError"}"#));
            }
        }

        Ok(FileCheckResult::Found) => {
            // If the file exists, serve it here!
            tracing::info!("New File View: {:?}", &filename.to_string());

            // Read the list of unsafe mimetypes from the config
            let mime_unsafe = &CONFIG.operations.unsafe_mimetype;

            // Get the mimetype of the file.
            let mut mimetype: String = db::get_mimetype(sqlconn, &filename).await.unwrap();

            // Use any instead of iterating over the collection manually
            // any is a method on an iterator that returns true if at least one element in the iterator satisfies a given predicate.
            // The predicate in this case is a closure that checks if the current mime type matches the unsafe mime type.
            if mime_unsafe
                .iter() // Iterate over the collection
                .any(|mime| mime.clone() == mimetype.clone())
            {
                // Check if any of the unsafe mime types match the given mimetype
                tracing::info!("Unsafe Extension Filtered: {:?}", mimetype);
                mimetype = String::from("text/plain"); // If there is a match, change the mimetype to text/plain
            }

            // Get the current unix time
            let accessed = SystemTime::now()
                .duration_since(SystemTime::UNIX_EPOCH)
                .unwrap()
                .as_secs() as i32;

            // Update the timestamp in the DB
            db::update_fileview(sqlconn, &filename, accessed)
                .await
                .unwrap_or_else(|err| {
                    tracing::error!(
                        "Failed to update fileview for file: {} error: {:?}",
                        &filename,
                        err
                    );
                    tracing::error!("soft-setting fileview to 0 for file: {}", &filename);
                    0
                });

            // Recalculate expiry for enginemode 1
            // we don't need filesize here, so it's 0.
            let filesize = 0;

            // TODO: we shouldn't need to recalculate the expiry on each view, even if it's in a different enginemode
            engine::calculate_expiry(sqlconn, filename.clone(), filesize).await;

            // Check if nginx sendfile is enabled, because we can skip this if it is.
            let nginxsendfile = CONFIG.server.nginx_sendfile;

            if nginxsendfile {
                // Add the X-Accel-Redirect header, allowing for faster file serving.
                // I was told to put a Path here, but I don't think it's really necessary, since it's converted to a HeaderValue anyway.
                let xsend = HeaderValue::try_from(format!("/files/{}", &filename)).unwrap();
                res.add_header("X-Accel-Redirect", xsend, true).unwrap();

                // Add the header to the response, which /may/ or /may not/ be modified by the safety check
                res.add_header("Content-Type", mimetype, true).unwrap();

                // Go through all the headers and print them out, just to check for now!
                tracing::debug!("response headers: {:?}", res.headers());

                // https://github.com/salvo-rs/salvo/issues/233
                res.status_code(StatusCode::OK);
            } else {
                // If nginx sendfile is disabled, we need to render the file directly

                let filepath = "files/".to_string() + &filename.to_string();
                let file = NamedFile::open(&filepath).await.unwrap();

                // Send the file to the response, and then send the response.
                file.send(&headers, res).await;

                // Add the header to the response, which /may/ or /may not/ be modified by the safety check
                res.add_header("Content-Type", mimetype, true).unwrap();

                // Go through all the headers and print them out, just to check for now!
                tracing::debug!("response headers: {:?}", res.headers());

                res.status_code(StatusCode::OK);
            }
        }
    };
}
